/******************************************************************************
 * Copyright (c) 2010-2011, Duane Merrill.  All rights reserved.
 * Copyright (c) 2011-2013, NVIDIA CORPORATION.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the NVIDIA CORPORATION nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NVIDIA CORPORATION BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************/

/******************************************************************************
 * Error handling utility routines
 ******************************************************************************/

#pragma once

#include <stdio.h>

B40C_NS_PREFIX

namespace b40c {
namespace util {


/**
 * Displays error message in accordance with debug mode
 */
template<int dummy>
cudaError_t B40CPerror(
	cudaError_t error,
	const char *message,
	const char *filename,
	int line,
	bool print)
{
	if (error && print) {
		fprintf(stderr, "[%s, %d] %s (CUDA error %d: %s)\n", filename, line, message, error, cudaGetErrorString(error));
		fflush(stderr);
	}
	return error;
}

template<int dummy>
cudaError_t B40CPerror(
	cudaError_t error,
	const char *message,
	const char *filename,
	int line)
{
	return B40CPerror<dummy>(error, message, filename, line, true);
}

/**
 * Checks and resets last CUDA error.  If set, displays last error message in accordance with debug mode.
 */
template<int dummy>
cudaError_t B40CPerror(
	const char *message,
	const char *filename,
	int line,
	bool print)
{
	cudaError_t error = cudaGetLastError();
	if (error && print) {

		fprintf(stderr, "[%s, %d] %s (CUDA error %d: %s)\n", filename, line, message, error, cudaGetErrorString(error));
		fflush(stderr);
	}
	return error;
}

template<int dummy>
cudaError_t B40CPerror(
	const char *message,
	const char *filename,
	int line)
{
	return B40CPerror<dummy>(message, filename, line, false);
}

/**
 * Displays error message in accordance with debug mode
 */
template<int dummy>
cudaError_t B40CPerror(cudaError_t error, bool print)
{
	if (error && print) {
		fprintf(stderr, "(CUDA error %d: %s)\n", error, cudaGetErrorString(error));
		fflush(stderr);
	}
	return error;
}

template<int dummy>
cudaError_t B40CPerror(cudaError_t error)
{
	return B40CPerror<dummy>(error, false);
}


/**
 * Checks and resets last CUDA error.  If set, displays last error message in accordance with debug mode.
 */
template<int dummy>
cudaError_t B40CPerror(bool print)
{
	cudaError_t error = cudaGetLastError();
	if (error && print) {
		fprintf(stderr, "(CUDA error %d: %s)\n", error, cudaGetErrorString(error));
		fflush(stderr);
	}
	return error;
}

template<int dummy>
cudaError_t B40CPerror(void)
{
	return B40CPerror<dummy>(false);
}

} // namespace util
} // namespace b40c

B40C_NS_POSTFIX

